import autopath
from py.test import raises

from tiramisu.setting import groups
from tiramisu.config import Config
from tiramisu.option import ChoiceOption, BoolOption, IntOption, FloatOption, \
    StrOption, OptionDescription
from tiramisu.error import PropertiesOptionError, ConflictError, SlaveError


def return_val():
    return 'val'


def return_list():
    return ['val', 'val']


def return_value(value):
    return value


def make_description():
    gcoption = ChoiceOption('name', 'GC name', ('ref', 'framework'), 'ref')
    gcdummy = BoolOption('dummy', 'dummy', default=False)
    objspaceoption = ChoiceOption('objspace', 'Object space',
                                  ('std', 'thunk'), 'std')
    booloption = BoolOption('bool', 'Test boolean option', default=True)
    intoption = IntOption('int', 'Test int option', default=0)
    intoption2 = IntOption('int', 'Test int option', default=0)
    floatoption = FloatOption('float', 'Test float option', default=2.3)
    stroption = StrOption('str', 'Test string option', default="abc")
    boolop = BoolOption('boolop', 'Test boolean option op', default=True)
    wantref_option = BoolOption('wantref', 'Test requires', default=False,
                                requires=((boolop, True, 'hidden'),))
    wantframework_option = BoolOption('wantframework', 'Test requires',
                                      default=False,
                                      requires=((boolop, True, 'hidden'),))
    gcgroup = OptionDescription('gc', '', [gcoption, gcdummy, floatoption, intoption2])
    descr = OptionDescription('constraints', '', [gcgroup, booloption, objspaceoption,
                              wantref_option, stroption,
                              wantframework_option,
                              intoption, boolop])
    return descr


def make_description_duplicates():
    gcoption = ChoiceOption('name', 'GC name', ('ref', 'framework'), 'ref')
    ## dummy 1
    gcdummy = BoolOption('dummy', 'dummy', default=False)
    objspaceoption = ChoiceOption('objspace', 'Object space',
                                  ('std', 'thunk'), 'std')
    booloption = BoolOption('bool', 'Test boolean option', default=True)
    intoption = IntOption('int', 'Test int option', default=0)
    floatoption = FloatOption('float', 'Test float option', default=2.3)
    stroption = StrOption('str', 'Test string option', default="abc")
    boolop = BoolOption('boolop', 'Test boolean option op', default=True)
    wantref_option = BoolOption('wantref', 'Test requires', default=False,
                                requires=((boolop, True, 'hidden'),))
    wantframework_option = BoolOption('wantframework', 'Test requires',
                                      default=False,
                                      requires=((boolop, True, 'hidden'),))
    # dummy2 (same path)
    gcdummy2 = BoolOption('dummy', 'dummy2', default=True)
    # dummy3 (same name)
    gcdummy3 = BoolOption('dummy', 'dummy2', default=True)
    gcgroup = OptionDescription('gc', '', [gcoption, gcdummy, gcdummy2, floatoption])
    descr = OptionDescription('constraints', '', [gcgroup, booloption, objspaceoption,
                              wantref_option, stroption,
                              wantframework_option,
                              intoption, boolop, gcdummy3])
    return descr


def test_identical_paths():
    """If in the schema (the option description) there is something that
    have the same name, an exection is raised
    """
    raises(ConflictError, "make_description_duplicates()")


def make_description2():
    gcoption = ChoiceOption('name', 'GC name', ['ref', 'framework'], 'ref')
    gcdummy = BoolOption('dummy', 'dummy', default=False)

    floatoption = FloatOption('float', 'Test float option', default=2.3)

    objspaceoption = ChoiceOption('objspace', 'Object space',
                                  ['std', 'thunk'], 'std')
    booloption = BoolOption('bool', 'Test boolean option', default=True)
    intoption = IntOption('int', 'Test int option', default=0)
    stroption = StrOption('str', 'Test string option', default="abc")
    # first multi
    boolop = BoolOption('boolop', 'Test boolean option op', default=True)
    boolop.enable_multi()
    wantref_option = BoolOption('wantref', 'Test requires', default=False,
                                requires=((boolop, True, 'hidden'),))
    # second multi
    wantframework_option = BoolOption('wantframework', 'Test requires',
                                      default=False,
                                      requires=((boolop, True, 'hidden'),))
    wantframework_option.enable_multi()

    gcgroup = OptionDescription('gc', '', [gcoption, gcdummy, floatoption])
    descr = OptionDescription('constraints', '', [gcgroup, booloption, objspaceoption,
                              wantref_option, stroption,
                              wantframework_option,
                              intoption, boolop])
    return descr


# FIXME: il faudra tester les validations sur les multis
#def test_multi_constraints():
#    "a multi in a constraint has to have the same length"
#    descr = make_description2()
#    cfg = Config(descr)
#    cfg.boolop = [True, True, False]
#    cfg.wantframework = [False, False, True]
#
#def test_multi_raise():
#    "a multi in a constraint has to have the same length"
#    # FIXME fusionner les deux tests, MAIS PROBLEME :
#    # il ne devrait pas etre necessaire de refaire une config
#    # si la valeur est modifiee une deuxieme fois ->
#    #raises(ConflictConfigError, "cfg.wantframework = [False, False, True]")
#    #   ExceptionFailure: 'DID NOT RAISE'
#    descr = make_description2()
#    cfg = Config(descr)
#    cfg.boolop = [True]
#    raises(ConflictConfigError, "cfg.wantframework = [False, False, True]")
# ____________________________________________________________
# adding dynamically new options description schema
#def test_newoption_add_in_descr():
#    descr = make_description()
#    newoption = BoolOption('newoption', 'dummy twoo', default=False)
#    descr.add_child(newoption)
#    config = Config(descr)
#    assert config.newoption == False

#def test_newoption_add_in_subdescr():
#    descr = make_description()
#    newoption = BoolOption('newoption', 'dummy twoo', default=False)
#    descr.gc.add_child(newoption)
#    config = Config(descr)
#    config.bool = False
#    assert config.gc.newoption == False

#def test_newoption_add_in_config():
#    descr = make_description()
#    config = Config(descr)
#    config.bool = False
#    newoption = BoolOption('newoption', 'dummy twoo', default=False)
#    descr.add_child(newoption)
#    config.cfgimpl_update()
#    assert config.newoption == False
# ____________________________________________________________


def make_description_requires():
    gcoption = ChoiceOption('name', 'GC name', ('ref', 'framework'), 'ref')
    gcdummy = BoolOption('dummy', 'dummy', default=False)

    floatoption = FloatOption('float', 'Test float option', default=2.3)

    objspaceoption = ChoiceOption('objspace', 'Object space',
                                  ('std', 'thunk'), 'std')
    booloption = BoolOption('bool', 'Test boolean option', default=True)
    intoption = IntOption('int', 'Test int option', default=0)
    stroption = StrOption('str', 'Test string option', default="abc",
                          requires=((intoption, 1, 'hidden'),))

    gcgroup = OptionDescription('gc', '', [gcoption, gcdummy, floatoption])
    descr = OptionDescription('constraints', '', [gcgroup, booloption, objspaceoption,
                              stroption, intoption])
    return descr


def test_hidden_if_in():
    descr = make_description_requires()
    cfg = Config(descr)
    setting = cfg.cfgimpl_get_settings()
    cfg.read_write()
    stroption = cfg.unwrap_from_path('str')
    assert not 'hidden' in setting[stroption]
    cfg.int = 1
    raises(PropertiesOptionError, "cfg.str")
    raises(PropertiesOptionError, 'cfg.str="uvw"')
    assert 'hidden' in setting[stroption]


def test_hidden_if_in_with_group():
    gcoption = ChoiceOption('name', 'GC name', ('ref', 'framework'), 'ref')
    gcdummy = BoolOption('dummy', 'dummy', default=False)

    floatoption = FloatOption('float', 'Test float option', default=2.3)

    objspaceoption = ChoiceOption('objspace', 'Object space',
                                  ('std', 'thunk'), 'std')
    booloption = BoolOption('bool', 'Test boolean option', default=True)
    intoption = IntOption('int', 'Test int option', default=0)
    stroption = StrOption('str', 'Test string option', default="abc")
    gcgroup = OptionDescription('gc', '', [gcoption, gcdummy, floatoption],
                                requires=((intoption, 1, 'hidden'),))
    descr = OptionDescription('constraints', '', [gcgroup, booloption,
                              objspaceoption, stroption, intoption])
    cfg = Config(descr)
    setting = cfg.cfgimpl_get_settings()
    cfg.read_write()
    assert not 'hidden' in setting[stroption]
    cfg.int = 1
    raises(PropertiesOptionError, "cfg.gc.name")


def test_disabled_with_group():
    gcoption = ChoiceOption('name', 'GC name', ('ref', 'framework'), 'ref')
    gcdummy = BoolOption('dummy', 'dummy', default=False)

    floatoption = FloatOption('float', 'Test float option', default=2.3)

    objspaceoption = ChoiceOption('objspace', 'Object space',
                                  ('std', 'thunk'), 'std')
    booloption = BoolOption('bool', 'Test boolean option', default=True)
    intoption = IntOption('int', 'Test int option', default=0)
    stroption = StrOption('str', 'Test string option', default="abc")
    gcgroup = OptionDescription('gc', '', [gcoption, gcdummy, floatoption],
                                requires=((intoption, 1, 'disabled'),))
    descr = OptionDescription('constraints', '', [gcgroup, booloption,
                              objspaceoption, stroption, intoption])
    cfg = Config(descr)
    cfg.read_write()
    assert cfg.gc.name
    cfg.int = 1
    raises(PropertiesOptionError, "cfg.gc.name")
#____________________________________________________________


def make_description_callback():
    gcoption = ChoiceOption('name', 'GC name', ('ref', 'framework'), 'ref')
    gcdummy = BoolOption('dummy', 'dummy')
    objspaceoption = ChoiceOption('objspace', 'Object space',
                                  ('std', 'thunk'), 'std')
    booloption = BoolOption('bool', 'Test boolean option', default=True)
    intoption = IntOption('int', 'Test int option', default=0)
    floatoption = FloatOption('float', 'Test float option', default=2.3)
    stroption = StrOption('str', 'Test string option', default="abc")
    boolop = BoolOption('boolop', 'Test boolean option op', default=True)
    wantref_option = BoolOption('wantref', 'Test requires', default=False,
                                requires=((boolop, True, 'hidden'),))
    wantframework_option = BoolOption('wantframework', 'Test requires',
                                      default=False,
                                      requires=((boolop, True, 'hidden'),))
    gcgroup = OptionDescription('gc', '', [gcoption, gcdummy, floatoption])
    descr = OptionDescription('constraints', '', [gcgroup, booloption, objspaceoption,
                              wantref_option, stroption,
                              wantframework_option,
                              intoption, boolop])
    return descr


def test_has_callback():
    descr = make_description_callback()
    # here the owner is 'default'
    config = Config(descr)
    setting = config.cfgimpl_get_settings()
    config.read_write()
    config.bool = False
    # because dummy has a callback
    dummy = config.unwrap_from_path('gc.dummy')
    setting.append('freeze')
    setting[dummy].append('frozen')
    raises(PropertiesOptionError, "config.gc.dummy = True")


def test_freeze_and_has_callback():
    descr = make_description_callback()
    config = Config(descr)
    setting = config.cfgimpl_get_settings()
    config.read_write()
    config.bool = False
    setting = config.cfgimpl_get_settings()
    setting.append('freeze')
    dummy = config.unwrap_from_path('gc.dummy')
    setting[dummy].append('frozen')
    raises(PropertiesOptionError, "config.gc.dummy = True")


def test_callback():
    val1 = StrOption('val1', "", callback=return_val)
    maconfig = OptionDescription('rootconfig', '', [val1])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1 == 'val'
    cfg.val1 = 'new-val'
    assert cfg.val1 == 'new-val'
    del(cfg.val1)
    assert cfg.val1 == 'val'


def test_callback_value():
    val1 = StrOption('val1', "", 'val')
    val2 = StrOption('val2', "", callback=return_value, callback_params={'': (('val1', False),)})
    maconfig = OptionDescription('rootconfig', '', [val1, val2])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1 == 'val'
    assert cfg.val2 == 'val'
    cfg.val1 = 'new-val'
    assert cfg.val1 == 'new-val'
    assert cfg.val2 == 'new-val'
    del(cfg.val1)
    assert cfg.val1 == 'val'
    assert cfg.val2 == 'val'


def test_callback_list():
    val1 = StrOption('val1', "", callback=return_list)
    maconfig = OptionDescription('rootconfig', '', [val1])
    cfg = Config(maconfig)
    cfg.read_write()
    raises(ValueError, "cfg.val1")


def test_callback_multi():
    val1 = StrOption('val1', "", callback=return_val, multi=True)
    maconfig = OptionDescription('rootconfig', '', [val1])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1 == ['val']
    cfg.val1 = ['new-val']
    assert cfg.val1 == ['new-val']
    cfg.val1.append('new-val2')
    assert cfg.val1 == ['new-val', 'new-val2']
    del(cfg.val1)
    assert cfg.val1 == ['val']


def test_callback_multi_value():
    val1 = StrOption('val1', "", ['val'], multi=True)
    val2 = StrOption('val2', "", multi=True, callback=return_value, callback_params={'': (('val1', False),)})
    maconfig = OptionDescription('rootconfig', '', [val1, val2])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1 == ['val']
    assert cfg.val2 == ['val']
    cfg.val1 = ['new-val']
    assert cfg.val1 == ['new-val']
    assert cfg.val2 == ['new-val']
    cfg.val1.append('new-val2')
    assert cfg.val1 == ['new-val', 'new-val2']
    assert cfg.val2 == ['new-val', 'new-val2']
    del(cfg.val1)
    assert cfg.val1 == ['val']
    assert cfg.val2 == ['val']


def test_callback_multi_list():
    val1 = StrOption('val1', "", callback=return_list, multi=True)
    maconfig = OptionDescription('rootconfig', '', [val1])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1 == ['val', 'val']
    cfg.val1 = ['new-val']
    assert cfg.val1 == ['new-val']
    cfg.val1.append('new-val2')
    assert cfg.val1 == ['new-val', 'new-val2']
    del(cfg.val1)
    assert cfg.val1 == ['val', 'val']


def test_callback_master_and_slaves_master():
    val1 = StrOption('val1', "", multi=True, callback=return_val)
    val2 = StrOption('val2', "", multi=True)
    interface1 = OptionDescription('val1', '', [val1, val2])
    interface1.impl_set_group_type(groups.master)
    maconfig = OptionDescription('rootconfig', '', [interface1])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1.val1 == ['val']
    cfg.val1.val1.append(None)
    assert cfg.val1.val1 == ['val', 'val']
    assert cfg.val1.val2 == [None, None]


def test_callback_master_and_slaves_master_list():
    val1 = StrOption('val1', "", multi=True, callback=return_list)
    val2 = StrOption('val2', "", multi=True)
    interface1 = OptionDescription('val1', '', [val1, val2])
    interface1.impl_set_group_type(groups.master)
    maconfig = OptionDescription('rootconfig', '', [interface1])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1.val1 == ['val', 'val']
    assert cfg.val1.val2 == [None, None]
    cfg.val1.val1.append(None)
    assert cfg.val1.val1 == ['val', 'val', None]
    assert cfg.val1.val2 == [None, None, None]
    del(cfg.val1.val1)
    assert cfg.val1.val1 == ['val', 'val']
    assert cfg.val1.val2 == [None, None]
    assert cfg.val1.val1.pop(1)
    assert cfg.val1.val1 == ['val']
    assert cfg.val1.val2 == [None]


def test_callback_master_and_slaves_slave():
    val1 = StrOption('val1', "", multi=True)
    val2 = StrOption('val2', "", multi=True, callback=return_val)
    interface1 = OptionDescription('val1', '', [val1, val2])
    interface1.impl_set_group_type(groups.master)
    maconfig = OptionDescription('rootconfig', '', [interface1])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1.val1 == []
    assert cfg.val1.val2 == []
    #
    cfg.val1.val1 = ['val1']
    assert cfg.val1.val1 == ['val1']
    assert cfg.val1.val2 == ['val']
    #
    cfg.val1.val1.append('val2')
    assert cfg.val1.val1 == ['val1', 'val2']
    assert cfg.val1.val2 == ['val', 'val']
    #
    cfg.val1.val1 = ['val1', 'val2', 'val3']
    assert cfg.val1.val1 == ['val1', 'val2', 'val3']
    assert cfg.val1.val2 == ['val', 'val', 'val']
    #
    cfg.val1.val1.pop(2)
    assert cfg.val1.val1 == ['val1', 'val2']
    assert cfg.val1.val2 == ['val', 'val']
    #
    cfg.val1.val2 = ['val2', 'val2']
    assert cfg.val1.val2 == ['val2', 'val2']
    #
    cfg.val1.val1.append('val3')
    assert cfg.val1.val2 == ['val2', 'val2', 'val']


def test_callback_master_and_slaves_slave_list():
    val1 = StrOption('val1', "", multi=True)
    val2 = StrOption('val2', "", multi=True, callback=return_list)
    interface1 = OptionDescription('val1', '', [val1, val2])
    interface1.impl_set_group_type(groups.master)
    maconfig = OptionDescription('rootconfig', '', [interface1])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1.val2 == []
    cfg.val1.val1 = ['val1', 'val2']
    assert cfg.val1.val1 == ['val1', 'val2']
    assert cfg.val1.val2 == ['val', 'val']
    cfg.val1.val1 = ['val1']
    #wrong len
    raises(SlaveError, 'cfg.val1.val2')


def test_callback_master_and_slaves_value():
    val1 = StrOption('val1', "", multi=True)
    val2 = StrOption('val2', "", multi=True, callback=return_value, callback_params={'': (('val1.val1', False),)})
    interface1 = OptionDescription('val1', '', [val1, val2])
    interface1.impl_set_group_type(groups.master)
    maconfig = OptionDescription('rootconfig', '', [interface1])
    cfg = Config(maconfig)
    cfg.read_write()
    assert cfg.val1.val1 == []
    assert cfg.val1.val2 == []
    #
    cfg.val1.val1 = ['val1']
    assert cfg.val1.val1 == ['val1']
    assert cfg.val1.val2 == ['val1']
    #
    cfg.val1.val1.append('val2')
    assert cfg.val1.val1 == ['val1', 'val2']
    assert cfg.val1.val2 == ['val1', 'val2']
    #
    cfg.val1.val1 = ['val1', 'val2', 'val3']
    assert cfg.val1.val1 == ['val1', 'val2', 'val3']
    assert cfg.val1.val2 == ['val1', 'val2', 'val3']
    #
    cfg.val1.val1.pop(2)
    assert cfg.val1.val1 == ['val1', 'val2']
    assert cfg.val1.val2 == ['val1', 'val2']
    #
    cfg.val1.val2 = ['val2', 'val2']
    assert cfg.val1.val2 == ['val2', 'val2']
    #
    cfg.val1.val1.append('val3')
    assert cfg.val1.val2 == ['val2', 'val2', 'val3']


